#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     David Herreros (dherreros@cnb.csic.es)
# *
# * National Centre for Biotechnology (CSIC), Spain
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import numpy as np
import shutil

from xmipp_metadata.metadata import XmippMetaData


# Change dir to correct path
package_path = os.path.abspath(os.path.dirname(__file__))
data_test_path = os.path.join(package_path, "data")
os.chdir(data_test_path)


# Clean output tests dir
for filename in os.listdir("test_outputs"):
    file_path = os.path.join("test_outputs", filename)
    try:
        if os.path.isfile(file_path) or os.path.islink(file_path):
            os.unlink(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path)
    except Exception as e:
        print('Failed to delete %s. Reason: %s' % (file_path, e))


# Read metadata
metadata = XmippMetaData("input_particles_two_sets.xmd")

# Read first image
image_1 = metadata.getMetaDataImage(0)

# Read second image
image_2 = metadata.getMetaDataImage(1)

# Get both images
images = metadata.getMetaDataImage([0, 1])

# Check images are close (Image 1)
error_image_1 = np.sqrt(np.mean((images[0] - image_1) ** 2))
print("Error for image 1: %.3f" % error_image_1)

# Check images are close (Image 2)
error_image_2 = np.sqrt(np.mean((images[1] - image_2) ** 2))
print("Error for image 2: %.3f" % error_image_2)

# Read metadata (Single set)
metadata_single = XmippMetaData("input_particles.xmd")

# Check images are close (Image 1)
image_1 = metadata_single.getMetaDataImage(0)
error_image_1 = np.sqrt(np.mean((images[0] - image_1) ** 2))
print("Error for image 1 (against single set): %.3f" % error_image_1)

# Check images are close (Image 2)
image_2 = metadata_single.getMetaDataImage(99)
error_image_2 = np.sqrt(np.mean((images[1] - image_2) ** 2))
print("Error for image 2 (against single set): %.3f" % error_image_2)

# Concat metadatas
metadata_twice = XmippMetaData("input_particles.xmd")
metadata_twice.appendMetaData(XmippMetaData("input_particles.xmd"))



# **************************************************************************
# *
# * Authors:     David Herreros (dherreros@cnb.csic.es)
# *
# * National Centre for Biotechnology (CSIC), Spain
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import math
from emtable import Table
import pandas as pd
from scipy.interpolate import RegularGridInterpolator
from scipy.spatial.transform import Rotation as R
from scipy.ndimage import affine_transform


# epsilon for testing whether a number is close to zero
_EPS = np.finfo(float).eps * 4.0

# axis sequences for Euler angles
_NEXT_AXIS = [1, 2, 0, 1]

# map axes strings to/from tuples of inner axis, parity, repetition, frame
_AXES2TUPLE = {
    'sxyz': (0, 0, 0, 0), 'sxyx': (0, 0, 1, 0), 'sxzy': (0, 1, 0, 0),
    'sxzx': (0, 1, 1, 0), 'syzx': (1, 0, 0, 0), 'syzy': (1, 0, 1, 0),
    'syxz': (1, 1, 0, 0), 'syxy': (1, 1, 1, 0), 'szxy': (2, 0, 0, 0),
    'szxz': (2, 0, 1, 0), 'szyx': (2, 1, 0, 0), 'szyz': (2, 1, 1, 0),
    'rzyx': (0, 0, 0, 1), 'rxyx': (0, 0, 1, 1), 'ryzx': (0, 1, 0, 1),
    'rxzx': (0, 1, 1, 1), 'rxzy': (1, 0, 0, 1), 'ryzy': (1, 0, 1, 1),
    'rzxy': (1, 1, 0, 1), 'ryxy': (1, 1, 1, 1), 'ryxz': (2, 0, 0, 1),
    'rzxz': (2, 0, 1, 1), 'rxyz': (2, 1, 0, 1), 'rzyz': (2, 1, 1, 1)}

_TUPLE2AXES = dict((v, k) for k, v in _AXES2TUPLE.items())


def emtable_2_pandas(file_name):
    """Convert an EMTable object to a Pandas dataframe to be used by XmippMetaData class"""

    # Read EMTable
    table = Table(fileName=file_name)

    # Init Pandas table
    pd_table = []

    # Iter rows and set data
    for row in table:
        row = row._asdict()
        for key, value in row.items():
            if isinstance(value, str) and not "@" in value:
                row[key] = value.replace(" ", ",")
        pd_table.append(pd.DataFrame([row]))

    return pd.concat(pd_table, ignore_index=True)


def fibonacci_sphere(samples):
    """
    Generate points on a unit sphere using the golden ratio-based Fibonacci lattice method.

    Args:
        samples (int): Number of points to generate.

    Returns:
        numpy.ndarray: Array of shape (samples, 3) containing 3D points on the sphere.
    """
    indices = np.arange(0, samples, dtype=float) + 0.5
    phi = np.arccos(1 - 2 * indices / samples)
    theta = np.pi * (1 + 5 ** 0.5) * indices

    x = np.sin(phi) * np.cos(theta)
    y = np.sin(phi) * np.sin(theta)
    z = np.cos(phi)

    return np.stack((z, y, x), axis=-1)


def fibonacci_hemisphere(n_points):
    n_points *= 2
    indices = np.arange(0, n_points, dtype=float) + 0.5

    phi = np.arccos(1 - 2 * indices / n_points)
    theta = np.pi * (1 + 5 ** 0.5) * indices

    # Mask to only take the upper hemisphere
    mask = (phi <= np.pi / 2)
    phi = phi[mask]
    theta = theta[mask]

    return theta, phi


def compute_rotations(theta, phi):
    # Rotation about the z-axis by theta
    Rz_theta = np.array([
        [np.cos(theta), -np.sin(theta), 0],
        [np.sin(theta), np.cos(theta), 0],
        [0, 0, 1]
    ])
    # Rotation about the y-axis by phi
    Ry_phi = np.array([
        [np.cos(phi), 0, np.sin(phi)],
        [0, 1, 0],
        [-np.sin(phi), 0, np.cos(phi)]
    ])
    # Combined rotation matrix
    return Ry_phi @ Rz_theta


# Fourier Slice Interpolator
class FourierInterpolator:
    def __init__(self, volume, pad):
        # Compute the Fourier transform of the volume
        self.size = volume.shape[0]
        self.pad = pad
        volume = np.pad(volume, int(0.25 * self.size * pad))
        self.pad_size = volume.shape[0]
        self.F = np.fft.fftshift(np.fft.fftn(np.fft.ifftshift(volume)))
        self.k = np.fft.fftshift(np.fft.fftfreq(volume.shape[0]))
        self.interpolator = RegularGridInterpolator(
            (self.k, self.k, self.k), self.F, bounds_error=False, fill_value=0
        )

    def get_slice(self, rot):
        # Define the grid points in each dimension
        z = np.fft.fftshift(np.fft.fftfreq(self.F.shape[0]))
        y = np.fft.fftshift(np.fft.fftfreq(self.F.shape[1]))
        x = np.fft.fftshift(np.fft.fftfreq(self.F.shape[2]))

        # Define the slice you want to interpolate in Fourier space
        z_slice_index = self.F.shape[0] // 2

        # Create a meshgrid for the slice in Fourier space
        Y, X = np.meshgrid(y, x, indexing='ij')
        Z = np.full_like(X, z[z_slice_index])

        # Flatten the coordinate arrays for transformation
        coords = np.array([X.ravel(), Y.ravel(), Z.ravel()])

        # Rotate the coordinates using the rotation matrix
        rotated_coords = np.dot(rot, coords)
        rotated_coords = np.vstack([rotated_coords[2, :], rotated_coords[1, :], rotated_coords[0, :]])

        # Get projection in real space
        projection = self.interpolator(rotated_coords.T).reshape(Z.shape)
        projection = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(projection))).real

        return projection.copy()


# Real space Slice Interpolator
class RealInterpolator:
    def __init__(self, volume):
        # Compute the Fourier transform of the volume
        self.size = volume.shape[0]
        self.pad_size = volume.shape[0]
        self.volume = volume
        self.k = np.fft.fftshift(np.fft.fftfreq(volume.shape[0]))
        self.interpolator = RegularGridInterpolator(
            (self.k, self.k, self.k), self.volume, bounds_error=False, fill_value=0
        )

    def get_slice(self, rot):
        """
        Rotate and prject a 3D volume using a given rotation matrix around its center.

        Args:
            volume (numpy.ndarray): 3D numpy array representing the volume.
            rotation_matrix (numpy.ndarray): 3x3 rotation matrix.

        Returns:
            numpy.ndarray: 2D projection.
        """
        # Volume shape
        volume_size = self.volume.shape

        # Define the grid points in each dimension
        z = np.fft.fftshift(np.fft.fftfreq(volume_size[0]))
        y = np.fft.fftshift(np.fft.fftfreq(volume_size[1]))
        x = np.fft.fftshift(np.fft.fftfreq(volume_size[2]))

        # Create a meshgrid of coordinates in the Fourier domain
        Z, Y, X = np.meshgrid(z, y, x, indexing='ij')

        # Flatten the coordinate arrays for transformation
        coords = np.array([X.ravel(), Y.ravel(), Z.ravel()])

        # Rotate the coordinates using the rotation matrix
        rotated_coords = np.dot(rot, coords)

        # Reshape the rotated coordinates back to the original shape
        rotated_Z = rotated_coords[2].reshape(Z.shape)
        rotated_Y = rotated_coords[1].reshape(Y.shape)
        rotated_X = rotated_coords[0].reshape(X.shape)

        # 4. Define the grid interpolator
        interpolator = RegularGridInterpolator((z, y, x), self.volume, method='linear', bounds_error=False,
                                               fill_value=0)

        # Interpolate the Fourier values at the rotated coordinates
        interpolated_values = interpolator((rotated_Z, rotated_Y, rotated_X))

        return np.sum(interpolated_values, axis=0).copy()


# Parallel Projection Computation using Joblib
def compute_projection(rot, interpolator):
    angles = -np.asarray(euler_from_matrix(rot, "szyz"))
    return interpolator.get_slice(np.linalg.inv(rot)), angles


def euler_from_matrix(matrix, axes='sxyz'):
    """Return Euler angles from rotation matrix for specified axis sequence.

    axes : One of 24 axis sequences as string or encoded tuple

    Note that many Euler angle triplets can describe one matrix.

    >>> R0 = euler_matrix(1, 2, 3, 'syxz')
    >>> al, be, ga = euler_from_matrix(R0, 'syxz')
    >>> R1 = euler_matrix(al, be, ga, 'syxz')
    >>> np.allclose(R0, R1)
    True
    >>> angles = (4*math.pi) * (np.random.random(3) - 0.5)
    >>> for axes in _AXES2TUPLE.keys():
    ...    R0 = euler_matrix(axes=axes, *angles)
    ...    R1 = euler_matrix(axes=axes, *euler_from_matrix(R0, axes))
    ...    if not np.allclose(R0, R1): print(axes, "failed")

    """
    try:
        firstaxis, parity, repetition, frame = _AXES2TUPLE[axes.lower()]
    except (AttributeError, KeyError):
        _TUPLE2AXES[axes]  # validation
        firstaxis, parity, repetition, frame = axes

    i = firstaxis
    j = _NEXT_AXIS[i + parity]
    k = _NEXT_AXIS[i - parity + 1]

    M = np.array(matrix, dtype=np.float64, copy=False)[:3, :3]
    if repetition:
        sy = math.sqrt(M[i, j] * M[i, j] + M[i, k] * M[i, k])
        if sy > _EPS:
            ax = math.atan2(M[i, j], M[i, k])
            ay = math.atan2(sy, M[i, i])
            az = math.atan2(M[j, i], -M[k, i])
        else:
            ax = math.atan2(-M[j, k], M[j, j])
            ay = math.atan2(sy, M[i, i])
            az = 0.0
    else:
        cy = math.sqrt(M[i, i] * M[i, i] + M[j, i] * M[j, i])
        if cy > _EPS:
            ax = math.atan2(M[k, j], M[k, k])
            ay = math.atan2(-M[k, i], cy)
            az = math.atan2(M[j, i], M[i, i])
        else:
            ax = math.atan2(-M[j, k], M[j, j])
            ay = math.atan2(-M[k, i], cy)
            az = 0.0

    if parity:
        ax, ay, az = -ax, -ay, -az
    if frame:
        ax, az = az, ax
    return ax, ay, az




# **************************************************************************
# *
# * Authors:     David Herreros (dherreros@cnb.csic.es)
# *
# * National Centre for Biotechnology (CSIC), Spain
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


from pathlib import Path

import numpy as np

from joblib import Parallel, delayed
from scipy.ndimage import shift

from skimage.transform import rescale, resize, warp
from skimage import filters
from skimage.measure import label
from skimage.morphology import opening, ball

import morphsnakes as ms

from scipy.ndimage.filters import gaussian_filter, median_filter
from scipy.spatial import ConvexHull, Delaunay

from .image_mrc import ImageMRC

from .image_spider import ImageSpider

from .image_em import ImageEM

from xmipp_metadata.utils import fibonacci_hemisphere, compute_rotations, FourierInterpolator, \
    compute_projection, RealInterpolator


class ImageHandler(object):
    '''
    Class to open several CryoEM image formats. Currently supported files include:
        - MRC files (supported trough mrcfile package)
        - Xmipp Spider files (STK and VOL)

    Currently, only reading operations are supported
    '''

    BINARIES = None
    DEBUG = False

    def __init__(self, binary_file=None):
        if binary_file:
            binary_file = str(binary_file)
            if ":mrcs" in binary_file:
                binary_file = binary_file.replace(":mrcs", "")
            elif ":mrc" in binary_file:
                binary_file = binary_file.replace(":mrc", "")

            self.binary_file = Path(binary_file)

            if self.binary_file.suffix == ".mrc" or self.binary_file.suffix == ".mrcs":
                self.BINARIES = ImageMRC(self.binary_file)
            elif self.binary_file.suffix == ".stk" or self.binary_file.suffix == ".vol" \
                    or self.binary_file.suffix == ".xmp" or self.binary_file.suffix == ".spi":
                self.BINARIES = ImageSpider(self.binary_file)
            elif self.binary_file.suffix == ".em" or self.binary_file.suffix == ".ems":
                self.BINARIES = ImageEM(self.binary_file)

    def __getitem__(self, item):
        if isinstance(self.BINARIES, ImageMRC):
            return self.BINARIES[item].copy()
        elif isinstance(self.BINARIES, ImageSpider):
            return self.BINARIES[item].copy()
        elif isinstance(self.BINARIES, ImageEM):
            return self.BINARIES.data[item].copy()
        else:
            return None

    def __len__(self):
        if isinstance(self.BINARIES, ImageSpider):
            return len(self.BINARIES)
        elif isinstance(self.BINARIES, ImageMRC):
            return len(self.BINARIES)
        elif isinstance(self.BINARIES, ImageEM):
            return len(self.BINARIES)
        else:
            return 0

    def __del__(self):
        if self.BINARIES is not None and isinstance(self.BINARIES, ImageSpider):
            self.BINARIES.close()
        if self.DEBUG:
            print("File closed succesfully!")

    def read(self, binary_file):
        '''
        Reading of a binary image file
            :param binary_file (string) --> Path to the binary file to be read
        '''
        if self.BINARIES:
            self.close()

        binary_file = str(binary_file)
        if ":mrcs" in binary_file:
            binary_file = binary_file.replace(":mrcs", "")
        elif ":mrc" in binary_file:
            binary_file = binary_file.replace(":mrc", "")

        self.binary_file = Path(binary_file)

        if self.binary_file.suffix == ".mrc" or self.binary_file.suffix == ".mrcs":
            self.BINARIES = ImageMRC(self.binary_file)
        elif self.binary_file.suffix == ".stk" or self.binary_file.suffix == ".vol" \
                or self.binary_file.suffix == ".xmp" or self.binary_file.suffix == ".spi":
            self.BINARIES = ImageSpider(self.binary_file)
        elif self.binary_file.suffix == ".em" or self.binary_file.suffix == ".ems":
            self.BINARIES = ImageEM(self.binary_file)

        return self

    def write(self, data, filename=None, overwrite=True, sr=1.0):
        if not overwrite and filename is None and len(self) != data.shape[0]:
            raise Exception("Cannot save file. Number of images "
                            "in new data is different. Please, set overwrite to True "
                            "if you are sure you want to do this.")

        filename = self.binary_file if filename is None else Path(filename)
        sr = 1.0 if sr == 0.0 else sr

        if filename.suffix == ".mrc" or filename.suffix == ".mrcs":
            ImageMRC().write(data, filename, overwrite=overwrite, sr=sr)
        elif filename.suffix == ".stk" or filename.suffix == ".vol" \
                or filename.suffix == ".xmp" or filename.suffix == ".spi":
            ImageSpider().write(data, filename, overwrite=overwrite, sr=sr)
        elif filename.suffix == ".em" or filename.suffix == ".ems":
            ImageEM().write(data, filename, overwrite=overwrite, sr=sr)

    def convert(self, orig_file, dest_file, overwrite=True):
        self.read(orig_file)
        data = self.getData()
        self.write(data, dest_file, sr=self.getSamplingRate(), overwrite=overwrite)

    def getData(self):
        return self[:]

    def getDimensions(self):
        if isinstance(self.BINARIES, ImageSpider):
            return np.asarray([self.BINARIES.header_info["n_slices"],
                               self.BINARIES.header_info["n_rows"],
                               self.BINARIES.header_info["n_columns"]])
        elif isinstance(self.BINARIES, ImageMRC):
            return np.asarray([self.BINARIES.header["nz"],
                               self.BINARIES.header["ny"],
                               self.BINARIES.header["nx"]])
        elif isinstance(self.BINARIES, ImageEM):
            return np.asarray([self.BINARIES.header["zdim"],
                               self.BINARIES.header["ydim"],
                               self.BINARIES.header["xdim"]])

    def getSamplingRate(self):
        if self.BINARIES is not None:
            return self.BINARIES.getSamplingRate()
        else:
            return None

    def setSamplingRate(self, input_file, sr):
        self.read(input_file)
        self.write(np.squeeze(self.getData()), sr=sr, overwrite=True)

    def scaleSplines(self, inputFn=None, outputFn=None, scaleFactor=None, finalDimension=None,
                     isStack=False, overwrite=True):
        if isinstance(inputFn, str):
            self.read(inputFn)
        data = np.squeeze(self.getData())

        if finalDimension is None:
            if isStack:
                aux = []
                for slice in data:
                    aux.append(rescale(slice, scaleFactor))
                data = np.asarray(aux)
            else:
                data = rescale(data, scaleFactor)
        else:
            # First check if dimesions are ok
            if isinstance(finalDimension, list) and len(finalDimension) != len(data.shape):
                raise ValueError(f"Resize dimensions do not match. You provided "
                                 f"{len(finalDimension)} dimensions, but data has "
                                 f"{len(data.shape)} dimensions")

            if isStack:
                aux = []
                if isinstance(finalDimension, int):
                    finalDimension = finalDimension * np.ones(len(data[0].shape))
                for slice in data:
                    aux.append(resize(slice, finalDimension))
                data = np.asarray(aux)
            else:
                if isinstance(finalDimension, int):
                    finalDimension = finalDimension * np.ones(len(data.shape))
                data = resize(data, finalDimension)

        scaleFactor = finalDimension[0] / data.shape[0] if scaleFactor is None else scaleFactor
        new_sr = self.getSamplingRate() / scaleFactor
        new_sr = new_sr if new_sr > 0.0 else 1.0

        if isinstance(outputFn, str):
            self.write(data, outputFn, sr=new_sr, overwrite=overwrite)
        else:
            return data

    def affineTransform(self, transformation, inputFn=None, outputFn=None, isStack=False, overwrite=True):
        if isinstance(inputFn, str):
            self.read(inputFn)
        data = np.squeeze(self.getData())

        # Transformation dim
        tr_dim = data.ndim if isStack else data.ndim + 1

        # Make transformation homogeneous (if not already)
        if transformation.shape[0] != tr_dim:
            transformation, aux = np.eye(tr_dim), transformation
            transformation[:-1, :-1] = aux

        # Get inverse transform
        inv_transformation = np.eye(tr_dim)
        inv_transformation[:-1, :-1] = np.linalg.inv(transformation[:-1, :-1])
        inv_transformation[:-1, -1] = -inv_transformation[:-1, :-1] @ transformation[:-1, -1]

        # Compute offset to rotate around image centre
        offset = 0.5 * np.asarray(data.shape[1:]) if isStack else 0.5 * np.asarray(data.shape)

        # Get grid coords
        if isStack:
            shape = [data.shape[1], data.shape[2]]
            Y, X = np.meshgrid(np.arange(shape[0]), np.arange(shape[1]))
            coords = np.array([X.ravel(), Y.ravel()])
        elif data.ndim == 2:
            shape = data.shape
            Y, X = np.meshgrid(np.arange(shape[0]), np.arange(shape[1]))
            coords = np.array([X.ravel(), Y.ravel()])
        else:
            shape = data.shape
            Z, Y, X = np.meshgrid(np.arange(shape[0]),
                                  np.arange(shape[1]),
                                  np.arange(shape[2]))
            coords = np.array([X.ravel(), Y.ravel(), Z.ravel()])
        coords = coords - offset[:, None]

        # Homogeneous coords
        coords = np.vstack([coords, np.zeros([1, coords.shape[1]])])

        # Get rotated coords
        rot_coords = inv_transformation @ coords

        # Coords in ZYX
        rot_coords = rot_coords[:-1, :]

        # Undo offset
        rot_coords = rot_coords + offset[:, None]

        # Warp coords
        coords = rot_coords.reshape([-1, ] + list(shape))

        if isStack:
            aux = []
            for slice in data:
                aux.append(warp(slice, coords))
            data = np.asarray(aux)
        else:
            data = warp(data, coords)

        sr = self.getSamplingRate() if self.getSamplingRate() > 0.0 else 1.0

        if isinstance(outputFn, str):
            self.write(data, outputFn, sr=sr, overwrite=overwrite)
        else:
            return data

    def createCircularMask(self, outputFile=None, boxSize=None, radius=None, center=None, is3D=True,
                           sr=1.0):
        if boxSize is None and radius is None:
            raise ValueError("At least boxSize or radius should be set.")

        boxSize = int(2 * radius) if boxSize is None else boxSize
        radius = 0.5 * boxSize if radius is None else radius

        if boxSize is not None:
            center = (0.5 * boxSize, 0.5 * boxSize)
        elif radius is not None:
            center = (radius, radius)

        # Create circular mask
        if is3D:
            Z, Y, X = np.ogrid[:boxSize, :boxSize, :boxSize]
            dist_from_center = np.sqrt((X - center[0]) ** 2 + (Y - center[1]) ** 2 + (Z - center[1]) ** 2)
        else:
            Y, X = np.ogrid[:boxSize, :boxSize]
            dist_from_center = np.sqrt((X - center[0]) ** 2 + (Y - center[1]) ** 2)

        mask = dist_from_center <= radius

        if isinstance(outputFile, str):
            self.write(mask, outputFile, overwrite=True, sr=sr)
        else:
            return mask

    def addNoise(self, input_file=None, output_file=None, std=1.0, avg=0.0, overwrite=True):
        if isinstance(input_file, str):
            self.read(input_file)
        data = np.squeeze(self.getData())
        noise = np.random.normal(loc=avg, scale=std, size=data.shape)
        data_noise = data + noise
        if isinstance(output_file, str):
            self.write(data_noise, output_file, overwrite=overwrite, sr=self.getSamplingRate())
        else:
            return output_file

    def __fillHull(self, image):
        """
        Compute the convex hull of the given binary image and
        return a mask of the filled hull.

        Adapted from:
        https://stackoverflow.com/a/46314485/162094
        This version is slightly (~40%) faster for 3D volumes,
        by being a little more stingy with RAM.
        """
        # (The variable names below assume 3D input,
        # but this would still work in 4D, etc.)

        assert (np.array(image.shape) <= np.iinfo(np.int16).max).all(), \
            f"This function assumes your image is smaller than {2 ** 15} in each dimension"

        points = np.argwhere(image).astype(np.int16)
        hull = ConvexHull(points)
        deln = Delaunay(points[hull.vertices])

        # Instead of allocating a giant array for all indices in the volume,
        # just iterate over the slices one at a time.
        idx_2d = np.indices(image.shape[1:], np.int16)
        idx_2d = np.moveaxis(idx_2d, 0, -1)

        idx_3d = np.zeros((*image.shape[1:], image.ndim), np.int16)
        idx_3d[:, :, 1:] = idx_2d

        mask = np.zeros_like(image, dtype=bool)
        for z in range(len(image)):
            idx_3d[:, :, 0] = z
            s = deln.find_simplex(idx_3d)
            mask[z, (s != -1)] = 1

        return mask

    def generateMask(self, iterations=150, smoothing=0, lambda1=1, lambda2=2, std=1, boxsize=128,
                     smoothStairEdges=False, keep_largest=True, dust_size=None, convex=True, applyThreshold=True,
                     threshold="otsu"):
        '''Generate automatically a binary protein mask based on a combination of snakes and
        Otsu method.
            :param int iterations: Number of iterations for computing the snake mask.
            :param int smoothing: Number of times the smoothing operator is applied per iteration.
                                  Reasonable values are around 0-4. Larger values lead to smoother
                                  segmentations.
            :param int lambda1: Weight parameter for the outer region. If `lambda1` is larger than
                                `lambda2`, the outer region will contain a larger range of values than
                                the inner region.
            :param int lambda2: Weight parameter for the inner region. If `lambda2` is larger than
                                `lambda1`, the inner region will contain a larger range of values than
                                the outer region.
            :param int std: Standard deviation to prefilter the map based on a Gaussian filter. Useful when
                            dealing with noisy maps
            :param int boxsize: Compute the snake mask for a map downsampled/upscaled to this size. We recommend
                                to downsample the map to a boxsize between 64px - 128px to improve performance.
                                The output size of the mask generated will keep the original map dimensions, even
                                if this parameter is used
            :param bool smoothStairEdges: Smooth the snake mask borders to avoid stair caise like borders.
                                          We recommend setting it to True if boxsized downsamples the original map.
            :param bool keep_largest: Keep the largest component only detected by the snakes mask. This will help the
                                      posterior Otsu thresholding step to find the appropriate threshold to segment the
                                      protein.
            :param str threshold: Threshold method use to improve the snakes masking. Valid options are:
                                  ["isodata", "li", "mean", "minimum", "otsu", "triangle", "yen"]
            :param float dust_size: Removes any connected component whose size is smaller than or equal to this value
            :param bool convex: Determines whether to create convex mask from the estimated morphSnake mask before
                                further processing it
            :param bool applyThreshold: Determines whether to apply a final thresholding step to generate a tight mask
                                        to the input image/volume
        '''
        # Read the data
        data = np.squeeze(self.getData())
        data_ori = data.copy()

        # Filter to remove noise (optional step)
        if std is not None:
            data = gaussian_filter(data, std)

        # Downscale to reduce execution times (optional but highly recommended)
        if boxsize is not None:
            ori_boxsize = data.shape[0]
            finalDimension = boxsize * np.ones(len(data.shape))
            data = resize(data, finalDimension)

        # Initialization for snakes
        init_ls = ms.circle_level_set(data.shape)

        # Generate snake mask
        acwe_ls1 = ms.morphological_chan_vese(data, iterations=iterations,
                                              init_level_set=init_ls, smoothing=smoothing,
                                              lambda1=lambda1, lambda2=lambda2)

        # Keep the largest component only
        if keep_largest:
            labels = label(acwe_ls1)
            assert (labels.max() != 0)  # assume at least 1 CC
            acwe_ls1 = labels == np.argmax(np.bincount(labels.flat)[1:]) + 1

        # Upscale mask (if downscaling was applied)
        if boxsize is not None:
            finalDimension = ori_boxsize * np.ones(len(data.shape))
            acwe_ls1 = resize(acwe_ls1.astype(bool), finalDimension).astype(np.float32)

        # Remove dust
        if dust_size is not None:
            acwe_ls1 = opening(acwe_ls1.astype(bool), ball(5))
            labels = label(acwe_ls1)
            assert (labels.max() != 0)  # assume at least 1 CC
            keep_regions = (np.argwhere(np.bincount(labels.flat)[1:] > dust_size) + 1)
            aux = np.zeros(acwe_ls1.shape)
            for lid in keep_regions:
                aux += labels == lid
            acwe_ls1 = aux.astype(np.float32)

        if smoothStairEdges:
            acwe_ls1 = (median_filter(acwe_ls1, size=5) >= 0.001).astype(np.float32)
        
        if convex:
            acwe_ls1 = self.__fillHull(acwe_ls1.astype(bool)).astype(float)

        if applyThreshold:
            data_ori = data_ori * acwe_ls1
            threshold_fun = getattr(filters, "threshold_" + threshold)
            acwe_ls1 = (data_ori >= threshold_fun(data_ori)).astype(np.float32)

        return acwe_ls1

    def generateProjections(self, num_projections, degrees=True, volume=None, n_jobs=8, pad=0, useFourier=True):
        """
        Generate 2D projections of a 3D volume uniformly over the projection sphere.

        Args:
            volume (numpy.ndarray): 3D numpy array representing the volume.
            num_projections (int): Number of projections to generate.

        Returns:
            projections (list of numpy.ndarray): List containing 2D projections.
            euler_angles (numpy.ndarray): Array of shape (num_projections, 3) containing
                                          Euler angles in ZYZ format for each projection.
        """
        # Read the data
        if volume is None:
            volume = np.squeeze(self.getData())

        theta, phi = fibonacci_hemisphere(num_projections)
        rots = np.stack([compute_rotations(t, p) for t, p in zip(theta, phi)], axis=0)

        if useFourier:
            interpolator = FourierInterpolator(volume, pad=pad)
        else:
            interpolator = RealInterpolator(volume)

        results = np.array(Parallel(n_jobs=n_jobs)(delayed(compute_projection)(rot, interpolator) for rot in rots),
                           dtype=object)
        projections, euler_angles = np.stack(results[:, 0], axis=0), np.stack(results[:, 1], axis=0)

        if degrees:
            euler_angles = np.rad2deg(euler_angles)

        return projections, euler_angles


    def close(self):
        '''
        Close the current binary file
        '''
        if isinstance(self.BINARIES, ImageSpider):
            self.BINARIES.close()



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     David Herreros (dherreros@cnb.csic.es)
# *
# * National Centre for Biotechnology (CSIC), Spain
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import shutil

import matplotlib.pyplot as plt
from scipy.spatial.transform import Rotation as R
import numpy as np
import time

from xmipp_metadata.image_handler import ImageHandler
from xmipp_metadata.metadata import XmippMetaData


# Change dir to correct path
package_path = os.path.abspath(os.path.dirname(__file__))
data_test_path = os.path.join(package_path, "data")
os.chdir(data_test_path)


# Create outputs dir
if not os.path.isdir("test_outputs"):
    os.mkdir("test_outputs")


# Clean output tests dir
for filename in os.listdir("test_outputs"):
    file_path = os.path.join("test_outputs", filename)
    try:
        if os.path.isfile(file_path) or os.path.islink(file_path):
            os.unlink(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path)
    except Exception as e:
        print('Failed to delete %s. Reason: %s' % (file_path, e))


# Read metadata
ih = ImageHandler("scaled_particles.stk")


# Get image with ImageHandler
img = ih[0]


# Write image (STK)
ih.write(img, filename=os.path.join("test_outputs", "test.stk"), sr=4.0)


# Write image (MRC)
ih.write(img, filename=os.path.join("test_outputs", "test.mrc"), sr=4.0)


# Raise error due to wrong overwrite
try:
    ih.write(img, overwrite=False)
except Exception as e:
    print("Error raised correctly!")


# Write image stack (STK)
img = ih[0:10]
ih.write(img, filename=os.path.join("test_outputs", "test_stack.stk"), sr=4.0)


# Write volume (VOL)
ih.write(img, filename=os.path.join("test_outputs", "test.vol"), sr=4.0)


# Convert (STK to MRCS)
ih.convert("scaled_particles.stk", os.path.join("test_outputs", "scaled_particles.mrcs"))


# Convert (STK to EMS)
ih.convert("scaled_particles.stk", os.path.join("test_outputs", "scaled_particles.ems"))


# Get dimensions (MRC)
ih.read(os.path.join("test_outputs", "scaled_particles.mrcs"))
dims_mrc = ih.getDimensions()


# Get dimensions (STK)
ih.read("scaled_particles.stk")
dims_stk = ih.getDimensions()


# Get dimensions (EMS)
ih.read(os.path.join("test_outputs", "scaled_particles.ems"))
dims_ems = ih.getDimensions()


# Scale stack (STK)
ih.scaleSplines("scaled_particles.stk",
                os.path.join("test_outputs", "test_stack_scaled.stk"),
                scaleFactor=2.0, isStack=True)


# Scale image (STK)
ih.scaleSplines(os.path.join("test_outputs", "test.stk"),
                os.path.join("test_outputs", "test_scaled.stk"),
                scaleFactor=2.0)


# Scale volume (VOL)
ih.scaleSplines("AK.vol",
                os.path.join("test_outputs", "test_scaled.vol"),
                finalDimension=[128, 128, 128])


# Scale stack (STK) with single int
ih.scaleSplines("scaled_particles.stk",
                os.path.join("test_outputs", "test_stack_scaled_int.stk"),
                finalDimension=128, isStack=True)


# Scale volume (VOL) with single int
ih.scaleSplines("AK.vol",
                os.path.join("test_outputs", "test_scaled_int.vol"),
                finalDimension=128)


# Scale volume (MRC) with single int
ih.scaleSplines("AK.vol",
                os.path.join("test_outputs", "test_scaled_int.mrc"),
                finalDimension=128)


# Scale volume (MRC) with single int (overwrite)
ImageHandler().scaleSplines(os.path.join("test_outputs", "test_scaled_int.mrc"),
                            os.path.join("test_outputs", "test_scaled_int.mrc"),
                            finalDimension=256, overwrite=True)


# Scale volume (VOL) with single int (overwrite)
ImageHandler().scaleSplines(os.path.join("test_outputs", "test_scaled_int.vol"),
                            os.path.join("test_outputs", "test_scaled_int.vol"),
                            finalDimension=256, overwrite=True)


# Check resize error due to wrong dimensions
# Raise error due to wrong overwrite
try:
    ih.scaleSplines("AK.vol",
                    os.path.join("test_outputs", "test_scaled_int.vol"),
                    finalDimension=[128, 128])
except Exception as e:
    print("Error raised correctly!")


# Create circular mask (image)
ih.createCircularMask(os.path.join("test_outputs", "mask_image.mrc"), boxSize=128, is3D=False,
                      sr=4.0)


# Create circular mask (volume)
ih.createCircularMask(os.path.join("test_outputs", "mask_vol.mrc"), boxSize=128, is3D=True,
                      sr=4.0)


# Warp stack (STK) (Rot 90º)
angle = 0.5 * np.pi
transform = np.eye(3)
transform[:-1, :-1] = np.asarray([[np.cos(angle), -np.sin(angle)],
                                  [np.sin(angle), np.cos(angle)]])
ImageHandler().affineTransform(inputFn="scaled_particles.stk",
                               outputFn=os.path.join("test_outputs", "test_stack_tr.stk",),
                               transformation=transform, isStack=True)


# Warp stack (VOL) (Rot 90º)
transform = np.eye(4)
transform[:-1, :-1] = R.from_euler("zyz", [0.0, 0.0, angle]).as_matrix()
ImageHandler().affineTransform(inputFn="AK.vol",
                               outputFn=os.path.join("test_outputs", "test_tr.vol"),
                               transformation=transform, isStack=False)


# Set sampling rate (VOL)
ImageHandler().setSamplingRate(os.path.join("test_outputs", "test_tr.vol"), sr=8.0)


# Add noise (VOL)
ImageHandler().addNoise(os.path.join("test_outputs", "test_scaled_int.vol"),
                        os.path.join("test_outputs", "test_tr.vol"),
                        avg=0.0, std=0.2, overwrite=True)


# Generate mask (VOL)
ih = ImageHandler(os.path.join("test_outputs", "test_tr.vol"))
start_time = time.time()
mask = ih.generateMask(iterations=50, boxsize=64, smoothStairEdges=False, dust_size=50)
end_time = time.time()
print(end_time - start_time)
ih.write(mask, os.path.join("test_outputs", "test_generated_mask.vol"), sr=ih.getSamplingRate())
ih.write(ih.getData() * mask, os.path.join("test_outputs", "test_generated_masked.vol"), sr=ih.getSamplingRate())


# Generate projections (Fourier)
ih = ImageHandler(os.path.join("AK.vol"))
volume = ih.scaleSplines(finalDimension=64)
print("Generating Fourier projections...")
start_time = time.time()
projections, angles = ih.generateProjections(500, degrees=True, pad=0, useFourier=True, volume=volume,
                                             n_jobs=20)
end_time = time.time()
print(end_time - start_time)
ImageHandler().write(projections, os.path.join("test_outputs", "projections_fourier.stk"))
ImageHandler().write(volume, os.path.join("test_outputs", "volume_to_poject_fourier.mrc"))
md = XmippMetaData(os.path.join("test_outputs", "projections_fourier.stk"), angles=angles)
md.setMetaDataColumns(np.ones(len(md)), "subtomo_labels")
md.write(os.path.join("test_outputs", "projections_fourier.xmd"), updateImagePaths=True)

# # Generate projections (Real)
ih = ImageHandler(os.path.join("AK.vol"))
volume = ih.scaleSplines(finalDimension=64)
print("Generating real projections...")
start_time = time.time()
projections, angles = ih.generateProjections(500, degrees=True, pad=0, useFourier=False, volume=volume,
                                             n_jobs=20)
end_time = time.time()
print(end_time - start_time)
ImageHandler().write(projections, os.path.join("test_outputs", "projections_real.stk"))
ImageHandler().write(volume, os.path.join("test_outputs", "volume_to_poject_real.mrc"))
md = XmippMetaData(os.path.join("test_outputs", "projections_real.stk"), angles=angles)
md.setMetaDataColumns(np.ones(len(md)), "subtomo_labels")
md.write(os.path.join("test_outputs", "projections_real.xmd"), updateImagePaths=True)



# **************************************************************************
# *
# * Authors:     David Herreros (dherreros@cnb.csic.es)
# *
# * National Centre for Biotechnology (CSIC), Spain
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import os

import mrcfile


class ImageMRC(object):
    '''
    Class to read an MRC file
    '''

    DEBUG = False

    def __init__(self, filename=None):
        if filename:
            self.read(filename)
        else:
            self.mrc_handle, self.header = None, None

    def __del__(self):
        '''
        Close the current file before deleting
        '''
        if self.DEBUG:
            print("File closed succesfully!")

    def __len__(self):
        return self.mrc_handle.header["nz"]

    def __iter__(self):
        '''
        Generator method to loop through all the images in the stack
        '''
        for image in self.data:
            yield image

    def __getitem__(self, item):
        return self.mrc_handle.data[item]

    def read(self, filename):
        '''
        Reads a given image
            :param filename (str) --> Image to be read
        '''
        try:
            self.mrc_handle = mrcfile.mmap(filename, mode='r+')
            self.header = self.mrc_handle.header
        except ValueError as e:
            print("MRC file header is not valid and raised the following error: ")
            print(e)
            print("We will try to read on permissive mode, and fix the header to recover your data")
            self.mrc_handle = mrcfile.mmap(filename, mode='r+', permissive=True)
            self.mrc_handle.update_header_from_data()
            self.header = self.mrc_handle.header

    def getSamplingRate(self):
        if self.mrc_handle is not None:
            return float(self.mrc_handle.voxel_size.x)
        else:
            return None

    def write(self, data, filename, overwrite=False, sr=1.0):
        sr = 1.0 if sr == 0.0 else sr

        if overwrite and os.path.isfile(filename):
            os.remove(filename)

        with mrcfile.new(filename, overwrite=overwrite) as mrc:
            mrc.set_data(data.astype(np.float32))
            mrc.voxel_size = sr



# **************************************************************************
# *
# * Authors:     David Herreros (dherreros@cnb.csic.es)
# *
# * National Centre for Biotechnology (CSIC), Spain
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np

import os

from pathlib import Path

import pandas as pd

import starfile

from xmipp_metadata.image_handler.image_handler import ImageHandler
from xmipp_metadata.utils import emtable_2_pandas


class XmippMetaData(object):
    '''
    Class to handle and Xmipp MetaData file (and its binaries) in Python

    Parameters:
        :param file_name (string - Optional) --> Path to metadata file
        :param readFrom (string - Optional) --> Can take values:
            - Auto: Determine automatically the best way to read the file
            - Pandas: Read the metadata file as a Pandas table
            - EMTable: Read the metadata file as a EMTable, which will be converted to Pandas later
    '''

    DEBUG = False
    DEFAULT_COLUMN_NAMES = ['anglePsi', 'angleRot', 'angleTilt', 'ctfVoltage', 'ctfDefocusU',
                            'ctfDefocusV', 'ctfDefocusAngle', 'ctfSphericalAberration', 'ctfQ0',
                            'enabled', 'flip', 'image', 'itemId', 'micrograph', 'micrographId',
                            'scoreByVariance', 'scoreByGiniCoeff', 'shiftX', 'shiftY', 'shiftZ',
                            'xcoor', 'ycoor']

    def __init__(self, file_name, readFrom="Auto", **kwargs):
        if file_name:
            if file_name.split(".")[-1] in ["xmd", "star"]:
                self.read(file_name, readFrom)
            elif file_name.split(".")[-1] in ["stk", "mrcs"]:  # Create new metadata from images
                # Fill metadata with images
                num_images = len(ImageHandler(file_name))
                angles = kwargs.pop("angles", np.zeros([num_images, 3]))
                shifts = kwargs.pop("shifts", np.zeros([num_images, 2]))
                res = {k: v for k, v in kwargs.items() if v is not None}
                COLUMN_DICT = {'anglePsi': angles[:, 2],
                               'angleRot': angles[:, 0],
                               'angleTilt': angles[:, 1],
                               'enabled': np.ones(num_images, dtype=int),
                               'image': [f"{id:06d}@{file_name}" for id in np.arange(1, num_images + 1, dtype=int)],
                               'itemId': np.arange(1, num_images + 1, dtype=int),
                               'shiftX': shifts[:, 0],
                               'shiftY': shifts[:, 1],
                               'shiftZ': np.zeros(num_images),
                               'ctfVoltage': np.zeros(num_images),
                               'ctfDefocusU': np.zeros(num_images),
                               'ctfDefocusV': np.zeros(num_images),
                               'ctfDefocusAngle': np.zeros(num_images),
                               'ctfSphericalAberration': np.zeros(num_images)}
                COLUMN_DICT.update(res)
                self.table = pd.DataFrame.from_dict(COLUMN_DICT)
                self.binaries = True

        else:
            self.table = pd.DataFrame(self.DEFAULT_COLUMN_NAMES)
            self.binaries = False

    def __len__(self):
        return self.table.shape[0]

    def __iter__(self):
        '''
        Iter through the rows in the metadata (generator method)
        '''
        for _, row in self.table.iterrows():
            yield row

    def __getitem__(self, item):
        extracted = self.table.loc[item]
        if hasattr(extracted, "to_numpy"):
            return extracted.to_numpy().copy()
        else:
            return extracted

    def __setitem__(self, key, value):
        self.table.loc[key] = value

    def read(self, file_name, readFrom="Auto"):
        '''
        Read a metadata file
            :param file_name (string) --> Path to metadata file
        '''
        if readFrom == "Auto":
            try:
                self.table = starfile.read(file_name)
            except ValueError:
                self.table = emtable_2_pandas(file_name)
        elif readFrom == "Pandas":
            self.table = starfile.read(file_name)
        elif readFrom == "EMTable":
            self.table = emtable_2_pandas(file_name)

        try:
            self.binaries = True
            _ = self.getMetaDataImage(0)
        except (FileNotFoundError, KeyError):
            self.binaries = False

        # Fill non-existing columns
        remain = set(self.DEFAULT_COLUMN_NAMES).difference(set(self.getMetaDataLabels()))
        for label in remain:
            self.table[label] = 0.0

    def write(self, filename, overwrite=True, updateImagePaths=False):
        '''
        Write current metadata to file
        '''
        # Filename path
        filename_path = Path(filename).resolve().parent

        # Image path
        def composeImageRelPath(image, relative_to):
            # Check if path has the form index@path
            try:
                index, file = image.split("@")
            except ValueError as e:
                index, file = "", image

            # Image absolute path
            if not os.path.isabs(file):
                file = os.path.abspath(file)

            # Get new relative path
            file = Path(file).resolve()
            file = os.path.relpath(file, start=relative_to)

            # Recompose path
            if index:
                image = index + "@" + file

            return image

        if updateImagePaths:
            for idx in range(len(self)):
                image = self.getMetadataItems(idx, "image")[0]
                image = composeImageRelPath(image, filename_path)
                self.setMetaDataItems(image, idx, "image")
        starfile.write(self.table, filename, overwrite=overwrite)

    def __del__(self):
        '''
        Closes the Metadata file and binaries to save memory
        '''
        if self.DEBUG:
            print("Binaries and MetaData closed successfully!")

    def shape(self):
        '''
        :returns: A tuple with the current metadata shape (rows, columns)
        '''
        return self.table.shape

    def getMetaDataRows(self, idx):
        '''
        Return a set of rows according to idx
            :parameter idx (list - int) --> Indices of the rows to be returned
            :returns The values stored in the desired rows as a Numpy array
        '''
        if isinstance(idx, (list, np.ndarray)) and len(idx) > 1:
            return self.table.iloc[idx].to_numpy().copy()
        else:
            return np.asarray([self.table.iloc[idx]])

    def setMetaDataRows(self, rows, idx):
        '''
        Set new values for metadata rows
        :param rows (Numpy array) --> New data to be set
        :param idx: (list - int) --> Rows indices to be set
        '''
        self.table.loc[idx, :] = rows
        
    def appendMetaDataRows(self, rows):
        self.table.loc[len(self.table.index)] = rows

    def appendMetaData(self, md):
        md.table["itemId"] = len(self) + md.table["itemId"]
        self.table = pd.concat([self.table, md.table], ignore_index=True)

    def getMetadataItems(self, rows_id, columns_id):
        '''
        Returns a slice of data in the metadata
            :param rows_id (list - int) --> Rows ids to be extracted
            :param columns_id (list - string, int) --> Columns names/indices to be extracted
            :return: sliced metadata as Numpy array
        '''
        if isinstance(rows_id, (list, np.ndarray)):
            return self.table.loc[rows_id, columns_id].to_numpy().copy()
        else:
            return np.asarray([self.table.loc[rows_id, columns_id]])

    def setMetaDataItems(self, items, rows_id, columns_id):
        '''
        Set new values for metadata columns
        :param items (Numpy array) --> New data to be set
        :param rows_id (list - int) --> Rows indices to be set
        :param columns_id (list - string, int) --> Columns names/indices to be set
        '''
        self.table.loc[rows_id, columns_id] = items

    def getMetaDataColumns(self, column_names):
        '''
        Return a set of rows according to idx
            :parameter column_names (list - string,int) --> Column names/indices to be returned
            :returns The values stored in the desired columns as a Numpy array
        '''
        return self.table.loc[:, column_names].to_numpy().copy()

    def setMetaDataColumns(self, columns, column_names):
        '''
        Set new values for metadata columns
        :param columns (Numpy array) --> New data to be set
        :param column_names: (list - string,int) --> Columns names/indices to be set
        '''
        self.table.loc[:, column_names] = columns

    def getMetaDataImage(self, row_id):
        '''
        Returns a set of images read from the metadata
            :param row_id (list - int) --> Row indices from where to read the images
            :returns: Images from metadata as Numpy array (N x Y x X)
        '''
        if self.binaries:
            images_rows = self.getMetadataItems(row_id, 'image')
            stack_id = {}
            stack_order = {}
            order_id = 0
            for row in images_rows:
                image_id, path = row.split('@') if "@" in row else (row_id, row)
                if path not in stack_id.keys():
                    stack_id[path] = [int(image_id) - 1, ]
                    stack_order[path] = [order_id, ]
                    order_id += 1
                else:
                    stack_id[path].append(int(image_id) - 1)
                    stack_order[path].append(order_id)
                    order_id += 1

            # Read binary file (if needed)
            images = []
            order = []
            for key, values in stack_id.items():
                order.append(np.asarray(stack_order[key]))
                images.append(ImageHandler(key)[values])
            order = np.hstack(order)
            images = np.squeeze(np.vstack(images))

            if order.size > 1:
                # Create an empty array of the same shape as the original array
                reordered_images = np.empty_like(images)

                # Reorder the original array based on the order vector
                reordered_images[order] = images

                return reordered_images
            else:
                return images
        else:
            print("Binaries not found...")

    def getMetaDataLabels(self):
        '''
        :returns: The metadata labels associated with the column in the current metadata
        '''
        return list(self.table.columns)

    def isMetaDataLabel(self, label):
        '''
        :returns: True or False depending on whether the metadata label is stored in the metadata
        '''
        return label in self.getMetaDataLabels()

    def concatenateMetadata(self, md):
        '''
        Concatenates a metadata file to the current metadata file
        '''
        if isinstance(md, str):
            md = XmippMetaData(md)

        self.table = pd.concat([self.table, md.table])



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
from xmipp_metadata import __version__

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='xmipp_metadata',  # Required
    version=__version__,  # Required
    description='Package to handle Xmipp Metadata and image binary data',  # Required
    long_description=long_description,  # Optional
    url='https://github.com/DavidHerreros/xmipp_metadata',  # Optional
    author='David Herreros',  # Optional
    author_email='dherreros@cnb.csic.es',  # Optional
    keywords='xmipp scipion cryoem imageprocessing scipion-3.0',  # Optional
    packages=find_packages(),
    install_requires=[requirements],
    # entry_points={'pyworkflow.plugin': 'myplugin = myplugin'},
    # package_data={  # Optional
    #    'myplugin': ['icon.png', 'protocols.conf'],
    # }
)



# **************************************************************************
# *
# * Authors:     David Herreros (dherreros@cnb.csic.es)
# *
# * National Centre for Biotechnology (CSIC), Spain
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import struct
import numpy as np
from pathlib import Path
import os


class ImageSpider(object):
    '''
    Class to read an STK image file generated with XMIPP
    '''

    HEADER_OFFSET = 1024
    FLOAT32_BYTES = 4
    TYPE = None
    DEBUG = False

    def __init__(self, filename=None):
        if filename:
            self.stk_handler = open(filename, "rb")
            self.header_info = self.read_header()
            self.IMG_BYTES = self.FLOAT32_BYTES * self.header_info["n_columns"] ** 2
        else:
            self.stk_handler, self.header_info, self.IMG_BYTES = None, None, None

    def __del__(self):
        '''
        Close the current file before deleting
        '''
        self.close()
        if self.DEBUG:
            print("File closed succesfully!")

    def __len__(self):
        if self.TYPE == "stack":
            return self.header_info["n_images"]
        elif self.TYPE == "volume":
            return self.header_info["n_slices"]

    def __iter__(self):
        '''
        Generator method to loop through all the images in the stack
        '''
        for iid in range(len(self)):
            yield self.read_image(iid)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.read_image(item)
        elif isinstance(item, list) or isinstance(item, np.ndarray):
            return np.stack([self.read_image(ii) for ii in item])
        elif isinstance(item, slice):
            start = item.start if item.start else 0
            stop = item.stop if item.stop else len(self)
            step = item.step if item.step else 1
            return np.stack([self.read_image(ii) for ii in range(start, stop, step)])

    def read(self, filename):
        '''
        Reads a given image
           :param filename (str) --> Image to be read
        '''
        self.stk_handler = open(filename, "rb")
        self.header_info = self.read_header()
        self.IMG_BYTES = self.FLOAT32_BYTES * self.header_info["n_columns"] ** 2

    def read_binary(self, start, end):
        '''
        Read bytes between start and end
            :param start (int) --> Start byte
            :param end (int) --> End byte
            :returns the bytes read
        '''
        self.seek(start)
        return self.stk_handler.read(end)

    def read_numpy(self, start, end):
        '''
        Read bytes between start and end as a Numpy array
            :param start (int) --> Start byte
            :param end (int) --> End byte
            :returns decoded bytes as Numpy array
        '''
        return np.frombuffer(self.read_binary(start, end), dtype=np.float32)

    def seek(self, pos):
        '''
        Move file pointer to a given position
            :param pos (int) --> Byte to move the pointer to
        '''
        self.stk_handler.seek(pos)

    def read_header(self):
        '''
        Reads the header of the current file as a dictionary
            :returns The current header as a dictionary
        '''
        header = self.read_numpy(0, self.HEADER_OFFSET)

        header = dict(img_size=int(header[1]), n_images=int(header[25]), offset=int(header[21]),
                      n_rows=int(header[1]), n_columns=int(header[11]), n_slices=int(header[0]),
                      sr=float(header[20]))

        self.TYPE = "stack" if header["n_images"] > 1 else "volume"

        return header

    def read_image(self, iid):
        '''
        Reads a given image in the stack according to its ID
            :param iid (int) --> Image id to be read
            :returns Image as Numpy array
        '''

        if self.TYPE == "stack":
            start = 2 * self.header_info["offset"] + iid * (self.IMG_BYTES + self.header_info["offset"])
        else:
            start = self.header_info["offset"] + iid * self.IMG_BYTES

        img_size = self.header_info["n_columns"]
        return self.read_numpy(start, self.IMG_BYTES).reshape([img_size, img_size])

    def write(self, data, filename=None, overwrite=False, sr=1.0):
        data = data.astype(np.float32)
        sr = 1.0 if sr == 0.0 else sr

        if overwrite and os.path.isfile(filename):
            os.remove(filename)

        if filename:
            mode = Path(filename).suffix
            fid = open(filename, "wb")
            if len(data.shape) == 3:
                # Write first header
                header = self.makeSpiderHeader(data, mode=mode, sr=sr)
                fid.writelines(header)

                # Write slices
                for slice in data:
                    if mode == ".stk":
                        header = self.makeSpiderHeader(slice[None, ...], mode=mode, sr=sr)
                        fid.writelines(header)
                    fid.writelines(slice)
            else:
                mode = ".vol"
                # Write image
                data = data[None, ...] if len(data.shape) == 2 else data
                header = self.makeSpiderHeader(data, mode=mode, sr=sr)
                fid.writelines(header)
                fid.writelines(data)
            fid.close()
        else:
            filename = self.stk_handler.name
            mode = Path(filename).suffix
            if data.shape[0] == len(self) or overwrite:
                fid = open(filename, "wb")
                if len(data.shape) == 3:
                    # Write first header
                    header = self.makeSpiderHeader(data, mode=mode, sr=sr)
                    fid.writelines(header)

                    # Write slices
                    for slice in data:
                        if mode == ".stk":
                            header = self.makeSpiderHeader(slice[None, ...], mode=mode, sr=sr)
                            fid.writelines(header)
                        fid.writelines(slice)
                else:
                    mode = ".vol"
                    # Write image
                    data = data[None, ...] if len(data.shape) == 2 else data
                    header = self.makeSpiderHeader(data, mode=mode, sr=sr)
                    fid.writelines(header)
                    fid.writelines(data)
                fid.close()
            else:
                raise Exception("Cannot save file. Number of images "
                                "in new data is different. Please, set overwrite to True "
                                "if you are sure you want to do this.")

    def makeSpiderHeader(self, im, mode, sr=1.0):
        n_slice, nsam, nrow = im.shape
        lenbyt = nsam * 4  # There are labrec records in the header
        labrec = int(1024 / lenbyt)
        if 1024 % lenbyt != 0:
            labrec += 1
        labbyt = labrec * lenbyt
        nvalues = int(labbyt / 4)
        if nvalues < 23:
            return []

        hdr = []
        for i in range(nvalues):
            hdr.append(0.0)

        # NB these are Fortran indices
        hdr[1] = float(n_slice) if mode == ".vol" else 1.0  # nslice (=1 for an image)
        hdr[2] = float(nrow)  # number of rows per slice
        hdr[3] = float(nrow)  # number of records in the image
        hdr[5] = 1.0 if mode == ".stk" else 3.0
        hdr[12] = float(nsam)  # number of pixels per line
        hdr[13] = float(labrec)  # number of records in file header
        hdr[21] = float(sr)  # sampling rate
        hdr[22] = float(labbyt)  # total number of bytes in header
        hdr[23] = float(lenbyt)  # record length in bytes
        hdr[24] = 2.0 if mode == ".stk" else 0.0
        hdr[25] = 1.0 if mode == ".stk" else 0.0
        hdr[26] = float(n_slice) if mode == ".stk" else 1.0  # nobjects (=1 for an image/vol)

        # adjust for Fortran indexing
        hdr = hdr[1:]
        hdr.append(0.0)
        # pack binary data into a string
        return [struct.pack("f", v) for v in hdr]

    def getSamplingRate(self):
        if self.header_info is not None:
            return self.header_info["sr"]
        else:
            return None

    def close(self):
        '''
        Closes the current file
        '''
        if self.stk_handler is not None:
            self.stk_handler.close()
